/*
    Copyright (C) 2004 Kor Nielsen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include "oss2jack.h"
#include <math.h>

/**
 * Retrieves the number of samples in the buffer
 * @param instance FIXME
 * @return FIXME
 */
int cb_get_count(struct circular_buffer* cbuf)
{
	if(cbuf->dma)
		return cbuf->buffer_size;
	
	int sample_count = cbuf->wptr - cbuf->rptr;
	if(sample_count < 0)
		sample_count += cbuf->buffer_size + 1;
	return sample_count;
}

int cb_get_available(struct circular_buffer* cbuf)
{
	if(cbuf->dma)
		return cbuf->buffer_size;
	
	int sample_count = cbuf->wptr - cbuf->rptr;
	if(sample_count < 0)
		sample_count += cbuf->buffer_size + 1;
	return cbuf->buffer_size - sample_count;
}


/**
 * FIXME
 * @param instance FIXME
 * @param samples FIXME
 * @param num_samples FIXME
 * @return FIXME
 */
int cb_read(struct circular_buffer* cbuf, float* samples, int num_samples)
{
	short* s;
	float* d;
	float* end;
	
	int max_samples = cb_get_count(cbuf);
	if(num_samples > max_samples)
		num_samples = max_samples;
	
	s = cbuf->rptr;
	d = samples;
	end = samples + num_samples;
	
	while(d < end)
	{
		(*d++) = (*s++) * ( 1.0f / 32768.0f);
		if(s >= cbuf->buffer + cbuf->buffer_size + 1 - cbuf->dma)
			s = cbuf->buffer;
	}
	return num_samples;
}


/**
 * FIXME
 * @param instance FIXME
 * @param samples FIXME
 * @param num_samples FIXME
 * @return FIXME
 */
int cb_read_short(struct circular_buffer* cbuf, short* samples, int num_samples)
{
	short* s;
	short* d;
	short* end;
	
	int max_samples = cb_get_count(cbuf);
	if(num_samples > max_samples)
		num_samples = max_samples;
	
	s = cbuf->rptr;
	d = samples;
	end = samples + num_samples;
	
	while(d < end)
	{
		(*d++) = (*s++);
		if(s >= cbuf->buffer + cbuf->buffer_size + 1 - cbuf->dma)
			s = cbuf->buffer;
	}
	cbuf->rptr = s;
	return num_samples;
}


int cb_read_stereo_mix(struct circular_buffer* cbuf, float* left, float* right, float vol_l, float vol_r, int num_samples)
{
	short* s;
	float* left_end;
	
	int max_samples = cb_get_count(cbuf);
	if(num_samples * 2 > max_samples)
		num_samples = max_samples / 2;
	
	s = cbuf->rptr;
	left_end = left + num_samples;
	
	while(left < left_end)
	{
		(*left++) += (*s++) * ( 1.0f / 32768.0f) * vol_l;
		if(s >= cbuf->buffer + cbuf->buffer_size + 1 - cbuf->dma)
			s = cbuf->buffer;
		
		(*right++) += (*s++) * ( 1.0f / 32768.0f) * vol_r;
		
		if(s >= cbuf->buffer + cbuf->buffer_size + 1 - cbuf->dma)
			s = cbuf->buffer;
	}
	
	cbuf->rptr = s;
	return num_samples;
}

int cb_read_mono_mix(struct circular_buffer* cbuf, float* left, float* right, float vol_l, float vol_r, int num_samples)
{
	short* s;
	float* left_end;
	
	int max_samples = cb_get_count(cbuf);
	if(num_samples > max_samples)
		num_samples = max_samples;
	
	s = cbuf->rptr;
	left_end = left + num_samples;
	
	while(left < left_end)
	{
		(*left++) += (*s) * ( 1.0f / 32768.0f) * vol_l;
		(*right++) += (*s++) * ( 1.0f / 32768.0f) * vol_r;
		if(s >= cbuf->buffer + cbuf->buffer_size + 1 - cbuf->dma)
			s = cbuf->buffer;
	}
	cbuf->rptr = s;
	return num_samples;
}

/**
 * FIXME
 * @param instance FIXME
 * @param num_samples FIXME
 */
int cb_skip(struct circular_buffer* cbuf, int num_samples)
{
	short* s;
	int total_samples = cb_get_count(cbuf);
	if(num_samples > total_samples)
		num_samples = total_samples;
		
	s = cbuf->rptr;
	s += num_samples;
	if(s >= cbuf->buffer + cbuf->buffer_size + 1 - cbuf->dma)
		s -= cbuf->buffer_size + 1 - cbuf->dma;
	
	cbuf->rptr = s;
	return num_samples;
}


int cb_write(struct circular_buffer* cbuf, float* samples, int count)
{
	int max_samples = cbuf->buffer_size - cb_get_count(cbuf);
	if(count > max_samples)
		count = max_samples;
		
	float* samples_end = samples + count;
	short* d_end = cbuf->buffer + cbuf->buffer_size + 1;
	short* d = cbuf->wptr;

	while(samples < samples_end)
	{
		(*d++) = (short) ((*samples++) * 32768.0f);
	
		if(d >= d_end)
			d = cbuf->buffer;
	}
	
	cbuf->wptr = d;
	return count;
}


int cb_write_short(struct circular_buffer* cbuf, short* s, int count)
{
	int max_samples = cbuf->buffer_size - cb_get_count(cbuf);
	if(count > max_samples)
		count = max_samples;
		
	short* s_end = s + count;
	short* d_end = cbuf->buffer + cbuf->buffer_size + 1;
	short* d = cbuf->wptr;

	while(s < s_end)
	{
		(*d++) = (*s++);
	
		if(d >= d_end)
			d = cbuf->buffer;
	}
	
	cbuf->wptr = d;
	return count;
}

int cb_write_stereo(struct circular_buffer* cbuf, float* left, float* right, float vol_l, float vol_r, int count)
{
	int max_samples = cbuf->buffer_size - cb_get_count(cbuf);
	if(count > max_samples * 2)
		count = max_samples / 2;
		
	float* left_end = left + count;
	short* d_end = cbuf->buffer + cbuf->buffer_size + 1;
	short* d = cbuf->wptr;

	while(left < left_end)
	{
		(*d++) = (short) ((*left++) * 32768.0f * vol_l);
		
		if(d >= d_end)
			d = cbuf->buffer;
			
		(*d++) = (short) ((*right++) * 32768.0f * vol_r);
		
		if(d >= d_end)
			d = cbuf->buffer;
	}
	
	cbuf->wptr = d;
	return count * 2;
}

int cb_write_mono(struct circular_buffer* cbuf, float* left, float* right, float vol_l, float vol_r, int count)
{
	int max_samples = cbuf->buffer_size - cb_get_count(cbuf);
	if(count > max_samples)
		count = max_samples;
		
	float* left_end = left + count;
	short* d_end = cbuf->buffer + cbuf->buffer_size + 1;
	short* d = cbuf->wptr;

	while(left < left_end)
	{
		(*d++) = (short) (((*left++) * vol_l + 
				   (*right++) * vol_r) * 32768.0f);
		
		if(d >= d_end)
			d = cbuf->buffer;
	}
	
	cbuf->wptr = d;
	return count;
}
